/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: carol@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: CMIServerExample.java 2439 2010-02-24 16:21:19Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.base.server;

import java.io.File;
import java.io.IOException;
import java.rmi.NoSuchObjectException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;

import org.ow2.cmi.admin.CMIAdmin;
import org.ow2.cmi.config.CMIConfig;
import org.ow2.cmi.config.CMIConfigurator;
import org.ow2.cmi.config.ProtocolConfig;
import org.ow2.cmi.controller.common.ClusterViewManager;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactory;
import org.ow2.util.execution.ExecutionResult;
import org.ow2.util.execution.IExecution;
import org.ow2.util.execution.helper.RunnableHelper;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * An example of CMI server
 * @author eyindanga
 *
 */
public class CMIServerExample {

	/**
	 * Logger.
	 */
	private static Log logger = LogFactory.getLog(CMIServerExample.class);

	/**
	 * Seconds is 1000 millis.
	 */
	private static final long SECONDS = 1000L;

	/**
	 * Current clusterview manager.
	 */
	private ClusterViewManager clusterViewManager = null;

	/**
	 * The current configuration.
	 */
	private CMIConfig<?> cmiConfig = null;

	/**
	 * Configuration directory.
	 */
	private File confDir = null;

	/**
	 * The configuration file.
	 */
	private String configFile = null;

	/**
	 * The jrmp registry.
	 */
	private Registry registry;

	/**
	 * CORBA registry.
	 */

	private Process tnameserv;
	/**
	 * @return the clusterViewManager
	 */
	public ClusterViewManager getClusterViewManager() {
		return clusterViewManager;
	}
	/**
	 * @param clusterViewManager the clusterViewManager to set
	 */
	public void setClusterViewManager(final ClusterViewManager clusterViewManager) {
		this.clusterViewManager = clusterViewManager;
	}
	/**
	 * @return the confDir
	 */
	public File getConfDir() {
		return confDir;
	}
	/**
	 * @param confDir the confDir to set
	 */
	public void setConfDir(final File confDir) {
		this.confDir = confDir;
	}
	/**
	 * @return the configFile
	 */
	public String getConfigFile() {
		return configFile;
	}
	/**
	 * @param configFile the configFile to set
	 */
	public void setConfigFile(final String configFile) {
		this.configFile = configFile;
	}
	/**
	 * @return the admin
	 */
	public CMIAdmin getAdmin() {
		return admin;
	}
	/**
	 * @param admin the admin to set
	 */
	public void setAdmin(final CMIAdmin admin) {
		this.admin = admin;
	}
	/**
	 * CMI Admin.
	 */
	private CMIAdmin admin;
	/**
	 * 
	 */
	private long killRegistryTimeout = 2000L;
	/**
	 * Default constructor.
	 */
	public CMIServerExample() {

	}
	/**
	 * Default constructor.
	 */
	public CMIServerExample(final String configDir, final String configFile) {
		this.confDir = new File(configDir);
		this.configFile = configFile;
	}

	/**
	 * server launcher.
	 */
	public final void start() {
		// Get the rmi port property
		final String confFileName = System.getProperty(
				ServerExampleProperty.CONF_FILE_URL_KEY.getPropertyName(),
				String.valueOf(ServerExampleProperty.DEFAULT_CONF_FILE_URL.getPropertyName()));
		// Use integer value
		logger.info("Starting CMI Server");
		try {
			IExecution<Void> registryExec = new IExecution<Void>() {

				public Void execute() throws Exception {


					File configFile = new File(confDir.getAbsolutePath(),
							confFileName);
					cmiConfig = CMIConfigurator.extractCMIConfig(configFile.toURL());
					CMIConfigurator.configure(ClusterViewManagerFactory.getFactory(), cmiConfig);
					// Gets the protocol.
					ProtocolConfig protocolConfig = (ProtocolConfig) cmiConfig.getJNDIConfig().getProtocolConfigs().toArray()[0];
					startRegistry(protocolConfig);

					clusterViewManager = ClusterViewManagerFactory.getFactory().create();
					clusterViewManager.start();
					admin = CMIAdmin.getCMIAdmin(clusterViewManager);
					return null;
				}

				/**
				 * Start local registry.
				 *
				 * @param localRegistryUrl
				 *            The registry Url
				 * @throws IOException
				 *             any.
				 */
				private void startRegistry(final ProtocolConfig protocolConfig)
				throws IOException {
					int port = getProtocolPort(protocolConfig
							.getLocalRegistryUrl());
					if (protocolConfig.getName().equals("jrmp")) {
						registry  = LocateRegistry.createRegistry(port);
						logger.info("jrmp registry started on port " + port);
					} else {
						tnameserv  = Runtime.getRuntime().exec(System.getProperty("java.home")
								+ System.getProperty("file.separator")
								+ "bin"
								+ System.getProperty("file.separator")
								+ "tnameserv -ORBInitialPort " + port);
						logger.info("iiop registry started on port " + port);

						//Thread that stops the iiop registry after a given timeout, whatever happens.
						new Thread("IIOP REGISTRY STOPPER") {

							@Override
							public void run() {
								try {
									String timeoutStr = System.getProperty(
											ServerExampleProperty.KILL_TIMEOUT_KEY.getPropertyName(),
											ServerExampleProperty.KILL_TIMEOUT_DEFAULT.getPropertyName());
									long killJvmTimeout = Long.parseLong(timeoutStr);
									logger.debug("The iiop regsitry will stop in " + killJvmTimeout / SECONDS + " seconds");
									Thread.sleep(killJvmTimeout);
								} catch (InterruptedException e) {
									logger.error("Cannot sleep");
								}
								// Exit JVM after timeout
								logger.debug("Timeout ! stopping the registry");
								if (tnameserv != null) {
									tnameserv.destroy();
								}
							}
						}.start();
					}
				}

				/**
				 * Gets the port for given protocol configuration.
				 *
				 * @param protocolConfig
				 *            The protocol's configuration.
				 * @return the registry port.
				 */
				private int getProtocolPort(final String protocolUrl) {
					int ind = protocolUrl.lastIndexOf(":");
					return Integer.parseInt(protocolUrl.substring(ind + 1));
				}
			};
			// Execute
			RunnableHelper <Void> nsHelper = new RunnableHelper <Void>();

			ExecutionResult <Void> nsStartExecResult = nsHelper.execute(CMIConfigurator.class.getClassLoader(), registryExec);

			if (nsStartExecResult.hasException()) {
				logger.error("Unable to start CMI ", nsStartExecResult.getException());
				return;
			}

		} catch (Exception e) {
			// TODO Auto-generated catch block
			logger.error("Unable to start the CMI registry ", e);
			return;
		}

	}
	/**
	 * @return the cmiConfig
	 */
	public CMIConfig<?> getCmiConfig() {
		return cmiConfig;
	}
	/**
	 * @param cmiConfig the cmiConfig to set
	 */
	public void setCmiConfig(final CMIConfig<?> cmiConfig) {
		this.cmiConfig = cmiConfig;
	}
	/**
	 * Stops the clusterview manager
	 * @throws NoSuchObjectException
	 */
	public void stop() throws NoSuchObjectException {
		clusterViewManager.stop();
		// TODO Auto-generated method stub
		if (registry != null) {
			UnicastRemoteObject.unexportObject(registry, true);
			registry = null;
		}else if (tnameserv != null) {
			//Stopping the iiop registry.
			tnameserv.destroy();
			tnameserv = null;
		}
		clusterViewManager = null;
		confDir = null;
	}

}
